@@ -55,6 +55,7 @@ typedef struct spice_migrate spice_migrate;
 #define FILE_XFER_CHUNK_SIZE (VD_AGENT_MAX_DATA_SIZE * 32)
 typedef struct SpiceFileXferTask {
     uint32_t                       id;
+    gboolean                       pending;
     GFile                          *file;
     SpiceMainChannel               *channel;
     GFileInputStream               *file_stream;
@@ -98,7 +99,7 @@ struct _SpiceMainChannelPrivate  {
     } display[MAX_DISPLAY];
     gint                        timer_id;
     GQueue                      *agent_msg_queue;
-    GList                       *file_xfer_task_list;
+    GHashTable                  *file_xfer_tasks;
     GSList                      *flushing;
 
     guint                       switch_host_delayed_id;
@@ -163,6 +164,8 @@ static void migrate_channel_event_cb(SpiceChannel *channel, SpiceChannelEvent ev
 static gboolean main_migrate_handshake_done(gpointer data);
 static void spice_main_channel_send_migration_handshake(SpiceChannel *channel);
 static void file_xfer_continue_read(SpiceFileXferTask *task);
+static void file_xfer_completed(SpiceFileXferTask *task, GError *error);
+static void file_xfer_flushed(SpiceMainChannel *channel, gboolean success);
 
 /* ------------------------------------------------------------------ */
 
@@ -205,6 +208,7 @@ static void spice_main_channel_init(SpiceMainChannel *channel)
 
     c = channel->priv = SPICE_MAIN_CHANNEL_GET_PRIVATE(channel);
     c->agent_msg_queue = g_queue_new();
+    c->file_xfer_tasks = g_hash_table_new(g_direct_hash, g_direct_equal);
 
     spice_main_channel_reset_capabilties(SPICE_CHANNEL(channel));
 }
@@ -312,6 +316,8 @@ static void spice_main_channel_finalize(GObject *obj)
 
     g_free(c->agent_msg_data);
     agent_free_msg_queue(SPICE_MAIN_CHANNEL(obj));
+    if (c->file_xfer_tasks)
+        g_hash_table_unref(c->file_xfer_tasks);
 
     if (G_OBJECT_CLASS(spice_main_channel_parent_class)->finalize)
         G_OBJECT_CLASS(spice_main_channel_parent_class)->finalize(obj);
@@ -327,22 +333,48 @@ static void spice_channel_iterate_write(SpiceChannel *channel)
 }
 
 /* main or coroutine context */
-static void spice_main_channel_reset(SpiceChannel *channel, gboolean migrating)
+static void spice_main_channel_reset_agent(SpiceMainChannel *channel)
 {
-    SpiceMainChannelPrivate *c = SPICE_MAIN_CHANNEL(channel)->priv;
+    SpiceMainChannelPrivate *c = channel->priv;
+    GError *error;
+    GList *tasks;
+    GList *l;
 
     c->agent_connected = FALSE;
     c->agent_caps_received = FALSE;
     c->agent_display_config_sent = FALSE;
-    c->agent_tokens = 0;
     c->agent_msg_pos = 0;
     g_free(c->agent_msg_data);
     c->agent_msg_data = NULL;
     c->agent_msg_size = 0;
 
+    tasks = g_hash_table_get_values(c->file_xfer_tasks);
+    for (l = tasks; l != NULL; l = l->next) {
+        SpiceFileXferTask *task = (SpiceFileXferTask *)l->data;
+
+        error = g_error_new(SPICE_CLIENT_ERROR, SPICE_CLIENT_ERROR_FAILED,
+                            "Agent connection closed");
+        file_xfer_completed(task, error);
+    }
+    g_list_free(tasks);
+    file_xfer_flushed(channel, FALSE);
+}
+
+/* main or coroutine context */
+static void spice_main_channel_reset(SpiceChannel *channel, gboolean migrating)
+{
+    SpiceMainChannelPrivate *c = SPICE_MAIN_CHANNEL(channel)->priv;
+
+    /* This is not part of reset_agent, since the spice-server expects any
+       pending multi-chunk messages to be completed by the client, even after
+       it has send an agent-disconnected msg as that is what the original
+       spicec did. Also see the TODO in server/reds.c reds_reset_vdp() */
+    c->agent_tokens = 0;
     agent_free_msg_queue(SPICE_MAIN_CHANNEL(channel));
     c->agent_msg_queue = g_queue_new();
 
+    spice_main_channel_reset_agent(SPICE_MAIN_CHANNEL(channel));
+
     SPICE_CHANNEL_CLASS(spice_main_channel_parent_class)->channel_reset(channel, migrating);
 }
 
@@ -866,17 +898,15 @@ static void file_xfer_flush_async(SpiceMainChannel *channel, GCancellable *cance
 static gboolean file_xfer_flush_finish(SpiceMainChannel *channel, GAsyncResult *result,
                                        GError **error)
 {
-    GSimpleAsyncResult *simple;
+    GSimpleAsyncResult *simple = (GSimpleAsyncResult *)result;
 
-    simple = (GSimpleAsyncResult *)result;
+    g_return_val_if_fail(g_simple_async_result_is_valid(result,
+        G_OBJECT(channel), file_xfer_flush_async), FALSE);
 
     if (g_simple_async_result_propagate_error(simple, error)) {
         return FALSE;
     }
 
-    g_return_val_if_fail(g_simple_async_result_is_valid(result,
-        G_OBJECT(channel), file_xfer_flush_async), FALSE);
-
     CHANNEL_DEBUG(channel, "flushed finished!");
     return g_simple_async_result_get_op_res_gboolean(simple);
 }
@@ -967,14 +997,15 @@ static void agent_msg_queue_many(SpiceMainChannel *channel, int type, const void
     g_warn_if_fail(out == NULL);
 }
 
-static int monitors_cmp(const void *p1, const void *p2)
+static int monitors_cmp(const void *p1, const void *p2, gpointer user_data)
 {
     const VDAgentMonConfig *m1 = p1;
     const VDAgentMonConfig *m2 = p2;
     double d1 = sqrt(m1->x * m1->x + m1->y * m1->y);
     double d2 = sqrt(m2->x * m2->x + m2->y * m2->y);
+    int diff = d1 - d2;
 
-    return d1 - d2;
+    return diff == 0 ? (char*)p1 - (char*)p2 : diff;
 }
 
 static void monitors_align(VDAgentMonConfig *monitors, int nmonitors)
@@ -988,7 +1019,7 @@ static void monitors_align(VDAgentMonConfig *monitors, int nmonitors)
 
     /* sort by distance from origin */
     sorted_monitors = g_memdup(monitors, nmonitors * sizeof(VDAgentMonConfig));
-    qsort(sorted_monitors, nmonitors, sizeof(VDAgentMonConfig), monitors_cmp);
+    g_qsort_with_data(sorted_monitors, nmonitors, sizeof(VDAgentMonConfig), monitors_cmp, NULL);
 
     /* super-KISS ltr alignment, feel free to improve */
     for (i = 0; i < nmonitors; i++) {
@@ -1006,8 +1037,9 @@ static void monitors_align(VDAgentMonConfig *monitors, int nmonitors)
         monitors[j].x = x;
         monitors[j].y = 0;
         x += monitors[j].width;
-        g_debug("#%d +%d+%d-%dx%d", j, monitors[j].x, monitors[j].y,
-                monitors[j].width, monitors[j].height);
+        if (monitors[j].width || monitors[j].height)
+            SPICE_DEBUG("#%d +%d+%d-%dx%d", j, monitors[j].x, monitors[j].y,
+                        monitors[j].width, monitors[j].height);
     }
     g_free(sorted_monitors);
 }
@@ -1064,10 +1096,10 @@ gboolean spice_main_send_monitor_config(SpiceMainChannel *channel)
             continue;
         }
         mon->monitors[j].depth  = c->display_color_depth ? c->display_color_depth : 32;
-        mon->monitors[j].width  = c->display[j].width;
-        mon->monitors[j].height = c->display[j].height;
-        mon->monitors[j].x = c->display[j].x;
-        mon->monitors[j].y = c->display[j].y;
+        mon->monitors[j].width  = c->display[i].width;
+        mon->monitors[j].height = c->display[i].height;
+        mon->monitors[j].x = c->display[i].x;
+        mon->monitors[j].y = c->display[i].y;
         CHANNEL_DEBUG(channel, "monitor config: #%d %dx%d+%d+%d @ %d bpp", j,
                       mon->monitors[j].width, mon->monitors[j].height,
                       mon->monitors[j].x, mon->monitors[j].y,
@@ -1288,14 +1320,40 @@ static void agent_clipboard_release(SpiceMainChannel *channel, guint selection)
     agent_msg_queue(channel, VD_AGENT_CLIPBOARD_RELEASE, msgsize, msg);
 }
 
+/* main context*/
+static gboolean timer_set_display(gpointer data)
+{
+    SpiceMainChannel *channel = data;
+    SpiceMainChannelPrivate *c = channel->priv;
+
+    c->timer_id = 0;
+    if (c->agent_connected)
+        spice_main_send_monitor_config(channel);
+
+    return FALSE;
+}
+
+/* any context  */
+static void update_display_timer(SpiceMainChannel *channel, guint seconds)
+{
+    SpiceMainChannelPrivate *c = channel->priv;
+
+    if (c->timer_id)
+        g_source_remove(c->timer_id);
+
+    c->timer_id = g_timeout_add_seconds(seconds, timer_set_display, channel);
+}
+
 /* coroutine context  */
 static void set_agent_connected(SpiceMainChannel *channel, gboolean connected)
 {
     SpiceMainChannelPrivate *c = channel->priv;
 
-    c->agent_connected = connected;
     SPICE_DEBUG("agent connected: %s", spice_yes_no(connected));
-    g_object_notify_main_context(G_OBJECT(channel), "agent-connected");
+    if (connected != c->agent_connected) {
+        c->agent_connected = connected;
+        g_object_notify_main_context(G_OBJECT(channel), "agent-connected");
+    }
 }
 
 /* coroutine context  */
@@ -1324,11 +1382,8 @@ static void agent_start(SpiceMainChannel *channel)
 /* coroutine context  */
 static void agent_stopped(SpiceMainChannel *channel)
 {
-    SpiceMainChannelPrivate *c = SPICE_MAIN_CHANNEL(channel)->priv;
-
-    set_agent_connected(channel, FALSE);
-    c->agent_caps_received = false;
-    c->agent_display_config_sent = false;
+    spice_main_channel_reset_agent(channel);
+    set_agent_connected(channel, FALSE); /* For notify */
     emit_main_context(channel, SPICE_MAIN_AGENT_UPDATE);
 }
 
@@ -1471,6 +1526,7 @@ static void main_handle_mouse_mode(SpiceChannel *channel, SpiceMsgIn *in)
 static void main_handle_agent_connected(SpiceChannel *channel, SpiceMsgIn *in)
 {
     agent_start(SPICE_MAIN_CHANNEL(channel));
+    update_display_timer(SPICE_MAIN_CHANNEL(channel), 0);
 }
 
 /* coroutine context */
@@ -1481,6 +1537,7 @@ static void main_handle_agent_connected_tokens(SpiceChannel *channel, SpiceMsgIn
 
     c->agent_tokens = msg->num_tokens;
     agent_start(SPICE_MAIN_CHANNEL(channel));
+    update_display_timer(SPICE_MAIN_CHANNEL(channel), 0);
 }
 
 /* coroutine context */
@@ -1489,18 +1546,6 @@ static void main_handle_agent_disconnected(SpiceChannel *channel, SpiceMsgIn *in
     agent_stopped(SPICE_MAIN_CHANNEL(channel));
 }
 
-static gint file_xfer_task_find(gconstpointer a, gconstpointer b)
-{
-    SpiceFileXferTask *task = (SpiceFileXferTask *)a;
-    uint32_t id = *(uint32_t *)b;
-
-    if (task->id == id) {
-        return 0;
-    }
-
-    return 1;
-}
-
 static void file_xfer_task_free(SpiceFileXferTask *task)
 {
     SpiceMainChannelPrivate *c;
@@ -1508,8 +1553,9 @@ static void file_xfer_task_free(SpiceFileXferTask *task)
     g_return_if_fail(task != NULL);
 
     c = task->channel->priv;
-    c->file_xfer_task_list = g_list_remove(c->file_xfer_task_list, task);
+    g_hash_table_remove(c->file_xfer_tasks, GUINT_TO_POINTER(task->id));
 
+    g_clear_object(&task->channel);
     g_clear_object(&task->file);
     g_clear_object(&task->file_stream);
     g_free(task);
@@ -1522,17 +1568,18 @@ static void file_xfer_close_cb(GObject      *object,
 {
     GSimpleAsyncResult *res;
     SpiceFileXferTask *task;
-    GInputStream *stream = G_INPUT_STREAM(object);
     GError *error = NULL;
 
-    stream = G_INPUT_STREAM(object);
     task = user_data;
 
-    g_input_stream_close_finish(stream, close_res, &error);
-    if (error) {
-        /* This error dont need to report to user, just print a log */
-        SPICE_DEBUG("close file error: %s", error->message);
-        g_clear_error(&error);
+    if (object) {
+        GInputStream *stream = G_INPUT_STREAM(object);
+        g_input_stream_close_finish(stream, close_res, &error);
+        if (error) {
+            /* This error dont need to report to user, just print a log */
+            SPICE_DEBUG("close file error: %s", error->message);
+            g_clear_error(&error);
+        }
     }
 
     /* Notify to user that files have been transferred or something error
@@ -1540,7 +1587,7 @@ static void file_xfer_close_cb(GObject      *object,
     res = g_simple_async_result_new(G_OBJECT(task->channel),
                                     task->callback,
                                     task->user_data,
-                                    file_xfer_continue_read);
+                                    spice_main_file_copy_async);
     if (task->error) {
         g_simple_async_result_take_error(res, task->error);
         g_simple_async_result_set_op_res_gboolean(res, FALSE);
@@ -1561,16 +1608,10 @@ static void file_xfer_data_flushed_cb(GObject *source_object,
     SpiceMainChannel *channel = (SpiceMainChannel *)source_object;
     GError *error = NULL;
 
+    task->pending = FALSE;
     file_xfer_flush_finish(channel, res, &error);
-
-    if (error != NULL) {
-        g_warning("failed to flush xfer queue: %s", error->message);
-        task->error = error;
-        g_input_stream_close_async(G_INPUT_STREAM(task->file_stream),
-                                   G_PRIORITY_DEFAULT,
-                                   task->cancellable,
-                                   file_xfer_close_cb,
-                                   task);
+    if (error || task->error) {
+        file_xfer_completed(task, error);
         return;
     }
 
@@ -1605,24 +1646,32 @@ static void file_xfer_read_cb(GObject *source_object,
     gssize count;
     GError *error = NULL;
 
+    task->pending = FALSE;
     count = g_input_stream_read_finish(G_INPUT_STREAM(task->file_stream),
                                        res, &error);
+    /* Check for pending earlier errors */
+    if (task->error) {
+        file_xfer_completed(task, error);
+        return;
+    }
+
     if (count > 0) {
         task->read_bytes += count;
         file_xfer_queue(task, count);
         file_xfer_flush_async(channel, task->cancellable,
                               file_xfer_data_flushed_cb, task);
-    } else {
-        /* Error or EOF, close the file */
-        if (error) {
-            task->error = error;
-        }
-        g_input_stream_close_async(G_INPUT_STREAM(task->file_stream),
-                                   G_PRIORITY_DEFAULT,
-                                   task->cancellable,
-                                   file_xfer_close_cb,
-                                   task);
+        task->pending = TRUE;
+    } else if (error) {
+        VDAgentFileXferStatusMessage msg = {
+            .id = task->id,
+            .result = VD_AGENT_FILE_XFER_STATUS_ERROR,
+        };
+        agent_msg_queue_many(task->channel, VD_AGENT_FILE_XFER_STATUS,
+                             &msg, sizeof(msg), NULL);
+        spice_channel_wakeup(SPICE_CHANNEL(task->channel), FALSE);
+        file_xfer_completed(task, error);
     }
+    /* else EOF, do nothing (wait for VD_AGENT_FILE_XFER_STATUS from agent) */
 }
 
 /* coroutine context */
@@ -1635,6 +1684,7 @@ static void file_xfer_continue_read(SpiceFileXferTask *task)
                               task->cancellable,
                               file_xfer_read_cb,
                               task);
+    task->pending = TRUE;
 }
 
 /* coroutine context */
@@ -1642,43 +1692,48 @@ static void file_xfer_handle_status(SpiceMainChannel *channel,
                                     VDAgentFileXferStatusMessage *msg)
 {
     SpiceMainChannelPrivate *c = channel->priv;
-    GList *l;
     SpiceFileXferTask *task;
+    GError *error = NULL;
 
-    l = g_list_find_custom(c->file_xfer_task_list, &msg->id,
-                           file_xfer_task_find);
 
-    g_return_if_fail(l != NULL);
-    task = l->data;
+    task = g_hash_table_lookup(c->file_xfer_tasks, GUINT_TO_POINTER(msg->id));
+    if (task == NULL) {
+        SPICE_DEBUG("cannot find task %d", msg->id);
+        return;
+    }
 
     SPICE_DEBUG("task %d received response %d", msg->id, msg->result);
 
     switch (msg->result) {
     case VD_AGENT_FILE_XFER_STATUS_CAN_SEND_DATA:
+        if (task->pending) {
+            error = g_error_new(SPICE_CLIENT_ERROR, SPICE_CLIENT_ERROR_FAILED,
+                           "transfer received CAN_SEND_DATA in pending state");
+            break;
+        }
         file_xfer_continue_read(task);
         return;
     case VD_AGENT_FILE_XFER_STATUS_CANCELLED:
-        SPICE_DEBUG("user removed task %d, result: %d", msg->id,
-                    msg->result);
-        task->error = g_error_new(SPICE_CLIENT_ERROR, SPICE_CLIENT_ERROR_FAILED,
-                                  "transfer is cancelled by spice agent");
+        error = g_error_new(SPICE_CLIENT_ERROR, SPICE_CLIENT_ERROR_FAILED,
+                            "transfer is cancelled by spice agent");
         break;
     case VD_AGENT_FILE_XFER_STATUS_ERROR:
-        task->error = g_error_new(SPICE_CLIENT_ERROR, SPICE_CLIENT_ERROR_FAILED,
-                                  "some errors occurred in the spice agent");
+        error = g_error_new(SPICE_CLIENT_ERROR, SPICE_CLIENT_ERROR_FAILED,
+                            "some errors occurred in the spice agent");
+        break;
+    case VD_AGENT_FILE_XFER_STATUS_SUCCESS:
+        if (task->pending)
+            error = g_error_new(SPICE_CLIENT_ERROR, SPICE_CLIENT_ERROR_FAILED,
+                                "transfer received success in pending state");
         break;
     default:
         g_warn_if_reached();
-        task->error = g_error_new(SPICE_CLIENT_ERROR, SPICE_CLIENT_ERROR_FAILED,
-                                  "unhandled status type: %u", msg->result);
+        error = g_error_new(SPICE_CLIENT_ERROR, SPICE_CLIENT_ERROR_FAILED,
+                            "unhandled status type: %u", msg->result);
         break;
     }
 
-    g_input_stream_close_async(G_INPUT_STREAM(task->file_stream),
-                               G_PRIORITY_DEFAULT,
-                               task->cancellable,
-                               file_xfer_close_cb,
-                               task);
+    file_xfer_completed(task, error);
 }
 
 /* coroutine context */
@@ -1836,6 +1891,8 @@ static void main_handle_agent_data(SpiceChannel *channel, SpiceMsgIn *in)
     guint8 *data;
     int len;
 
+    g_warn_if_fail(c->agent_connected);
+
     /* shortcut to avoid extra message allocation & copy if possible */
     if (c->agent_msg_pos == 0) {
         VDAgentMessage *msg;
@@ -2295,19 +2352,6 @@ static void spice_main_handle_msg(SpiceChannel *channel, SpiceMsgIn *msg)
         g_return_if_reached();
 }
 
-/* system context*/
-static gboolean timer_set_display(gpointer data)
-{
-    SpiceChannel *channel = data;
-    SpiceMainChannelPrivate *c = SPICE_MAIN_CHANNEL(channel)->priv;
-
-    c->timer_id = 0;
-    if (c->agent_connected)
-        spice_main_send_monitor_config(SPICE_MAIN_CHANNEL(channel));
-
-    return false;
-}
-
 /**
  * spice_main_agent_test_capability:
  * @channel:
@@ -2330,19 +2374,25 @@ gboolean spice_main_agent_test_capability(SpiceMainChannel *channel, guint32 cap
 }
 
 /**
- * spice_main_set_display:
+ * spice_main_update_display:
  * @channel:
  * @id: display ID
  * @x: x position
  * @y: y position
  * @width: display width
  * @height: display height
+ * @update: if %TRUE, update guest resolution after 1sec.
  *
- * Notify the guest of screen resolution change. The notification is
- * sent 1 second later, if no further changes happen.
+ * Update the display @id resolution.
+ *
+ * If @update is %TRUE, the remote configuration will be updated too
+ * after 1 second without further changes. You can send when you want
+ * without delay the new configuration to the remote with
+ * spice_main_send_monitor_config()
  **/
-void spice_main_set_display(SpiceMainChannel *channel, int id,
-                            int x, int y, int width, int height)
+void spice_main_update_display(SpiceMainChannel *channel, int id,
+                               int x, int y, int width, int height,
+                               gboolean update)
 {
     SpiceMainChannelPrivate *c;
 
@@ -2362,10 +2412,26 @@ void spice_main_set_display(SpiceMainChannel *channel, int id,
     c->display[id].width  = width;
     c->display[id].height = height;
 
-    if (c->timer_id) {
-        g_source_remove(c->timer_id);
-    }
-    c->timer_id = g_timeout_add_seconds(1, timer_set_display, channel);
+    if (update)
+        update_display_timer(channel, 1);
+}
+
+/**
+ * spice_main_set_display:
+ * @channel:
+ * @id: display ID
+ * @x: x position
+ * @y: y position
+ * @width: display width
+ * @height: display height
+ *
+ * Notify the guest of screen resolution change. The notification is
+ * sent 1 second later, if no further changes happen.
+ **/
+void spice_main_set_display(SpiceMainChannel *channel, int id,
+                            int x, int y, int width, int height)
+{
+    spice_main_update_display(channel, id, x, y, width, height, TRUE);
 }
 
 /**
@@ -2547,23 +2613,34 @@ void spice_main_set_display_enabled(SpiceMainChannel *channel, int id, gboolean
         c->display[id].enabled = enabled;
     }
 
-    if (c->timer_id) {
-        g_source_remove(c->timer_id);
-    }
-    c->timer_id = g_timeout_add_seconds(1, timer_set_display, channel);
+    update_display_timer(channel, 1);
 }
 
-static void file_xfer_failed(SpiceFileXferTask *task, GError *error)
+static void file_xfer_completed(SpiceFileXferTask *task, GError *error)
 {
-    SPICE_DEBUG("File %s xfer failed: %s",
-                g_file_get_path(task->file), error->message);
+    /* In case of multiple errors we only report the first error */
+    if (task->error)
+        g_clear_error(&error);
+    if (error) {
+        SPICE_DEBUG("File %s xfer failed: %s",
+                    g_file_get_path(task->file), error->message);
+        task->error = error;
+    }
+
+    if (task->pending)
+        return;
+
+    if (!task->file_stream) {
+        file_xfer_close_cb(NULL, NULL, task);
+        return;
+    }
 
-    task->error = error;
     g_input_stream_close_async(G_INPUT_STREAM(task->file_stream),
                                G_PRIORITY_DEFAULT,
                                task->cancellable,
                                file_xfer_close_cb,
                                task);
+    task->pending = TRUE;
 }
 
 static void file_xfer_info_async_cb(GObject *obj, GAsyncResult *res, gpointer data)
@@ -2577,10 +2654,10 @@ static void file_xfer_info_async_cb(GObject *obj, GAsyncResult *res, gpointer da
     gsize /*msg_size*/ data_len;
     gchar *string;
     SpiceFileXferTask *task = (SpiceFileXferTask *)data;
-    SpiceMainChannelPrivate *c = task->channel->priv;
 
+    task->pending = FALSE;
     info = g_file_query_info_finish(file, res, &error);
-    if (error)
+    if (error || task->error)
         goto failed;
 
     task->file_size =
@@ -2602,9 +2679,6 @@ static void file_xfer_info_async_cb(GObject *obj, GAsyncResult *res, gpointer da
         goto failed;
 
     /* Create file-xfer start message */
-    CHANNEL_DEBUG(task->channel, "Insert a xfer task:%d to task list", task->id);
-    c->file_xfer_task_list = g_list_append(c->file_xfer_task_list, task);
-
     msg.id = task->id;
     agent_msg_queue_many(task->channel, VD_AGENT_FILE_XFER_START,
                          &msg, sizeof(msg),
@@ -2614,7 +2688,7 @@ static void file_xfer_info_async_cb(GObject *obj, GAsyncResult *res, gpointer da
     return;
 
 failed:
-    file_xfer_failed(task, error);
+    file_xfer_completed(task, error);
 }
 
 static void file_xfer_read_async_cb(GObject *obj, GAsyncResult *res, gpointer data)
@@ -2623,9 +2697,10 @@ static void file_xfer_read_async_cb(GObject *obj, GAsyncResult *res, gpointer da
     SpiceFileXferTask *task = (SpiceFileXferTask *)data;
     GError *error = NULL;
 
+    task->pending = FALSE;
     task->file_stream = g_file_read_finish(file, res, &error);
-    if (error) {
-        file_xfer_failed (task, error);
+    if (error || task->error) {
+        file_xfer_completed(task, error);
         return;
     }
 
@@ -2636,6 +2711,7 @@ static void file_xfer_read_async_cb(GObject *obj, GAsyncResult *res, gpointer da
                             task->cancellable,
                             file_xfer_info_async_cb,
                             task);
+    task->pending = TRUE;
 }
 
 static void file_xfer_send_start_msg_async(SpiceMainChannel *channel,
@@ -2647,14 +2723,13 @@ static void file_xfer_send_start_msg_async(SpiceMainChannel *channel,
                                            GAsyncReadyCallback callback,
                                            gpointer user_data)
 {
+    SpiceMainChannelPrivate *c = channel->priv;
     SpiceFileXferTask *task;
     static uint32_t xfer_id;    /* Used to identify task id */
 
-    xfer_id = (xfer_id > UINT32_MAX) ? 0 : xfer_id;
-
     task = spice_malloc0(sizeof(SpiceFileXferTask));
     task->id = ++xfer_id;
-    task->channel = channel;
+    task->channel = g_object_ref(channel);
     task->file = g_object_ref(file);
     task->flags = flags;
     task->cancellable = cancellable;
@@ -2663,12 +2738,15 @@ static void file_xfer_send_start_msg_async(SpiceMainChannel *channel,
     task->callback = callback;
     task->user_data = user_data;
 
+    CHANNEL_DEBUG(task->channel, "Insert a xfer task:%d to task list", task->id);
+    g_hash_table_insert(c->file_xfer_tasks, GUINT_TO_POINTER(task->id), task);
+
     g_file_read_async(file,
                       G_PRIORITY_DEFAULT,
                       cancellable,
                       file_xfer_read_async_cb,
                       task);
-
+    task->pending = TRUE;
 }
 
 /**
@@ -2754,7 +2832,8 @@ gboolean spice_main_file_copy_finish(SpiceMainChannel *channel,
     GSimpleAsyncResult *simple;
 
     g_return_val_if_fail(SPICE_IS_MAIN_CHANNEL(channel), FALSE);
-    g_return_val_if_fail(result != NULL, FALSE);
+    g_return_val_if_fail(g_simple_async_result_is_valid(result,
+        G_OBJECT(channel), spice_main_file_copy_async), FALSE);
 
     simple = (GSimpleAsyncResult *)result;
 
