@@ -652,7 +652,9 @@ static void spice_session_class_init(SpiceSessionClass *klass)
 {
     GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
 
+#if GLIB_CHECK_VERSION(2, 26, 0)
     _wocky_http_proxy_get_type();
+#endif
 
     gobject_class->dispose      = spice_session_dispose;
     gobject_class->finalize     = spice_session_finalize;
@@ -1177,13 +1179,17 @@ SpiceSession *spice_session_new_from_session(SpiceSession *session)
                                                     NULL));
     SpiceSessionPrivate *c = copy->priv, *s = session->priv;
 
-    g_warn_if_fail (c->host == NULL);
-    g_warn_if_fail (c->tls_port == NULL);
-    g_warn_if_fail (c->password == NULL);
-    g_warn_if_fail (c->ca_file == NULL);
-    g_warn_if_fail (c->ciphers == NULL);
-    g_warn_if_fail (c->cert_subject == NULL);
-    g_warn_if_fail (c->pubkey == NULL);
+    g_clear_object(&c->proxy);
+
+    g_warn_if_fail(c->host == NULL);
+    g_warn_if_fail(c->tls_port == NULL);
+    g_warn_if_fail(c->password == NULL);
+    g_warn_if_fail(c->ca_file == NULL);
+    g_warn_if_fail(c->ciphers == NULL);
+    g_warn_if_fail(c->cert_subject == NULL);
+    g_warn_if_fail(c->pubkey == NULL);
+    g_warn_if_fail(c->pubkey == NULL);
+    g_warn_if_fail(c->proxy == NULL);
 
     g_object_get(session,
                  "host", &c->host,
@@ -1204,6 +1210,8 @@ SpiceSession *spice_session_new_from_session(SpiceSession *session)
     c->client_provided_sockets = s->client_provided_sockets;
     c->protocol = s->protocol;
     c->connection_id = s->connection_id;
+    if (s->proxy)
+        c->proxy = g_object_ref(s->proxy);
 
     return copy;
 }
@@ -1637,7 +1645,7 @@ struct spice_open_host {
     int port;
     GCancellable *cancellable;
     GError *error;
-    GSocket *socket;
+    GSocketConnection *connection;
     GSocketClient *client;
 };
 
@@ -1653,11 +1661,9 @@ static void socket_client_connect_ready(GObject *source_object, GAsyncResult *re
     if (connection == NULL)
         goto end;
 
-    open_host->socket = g_socket_connection_get_socket(connection);
-    g_object_ref(open_host->socket);
+    open_host->connection = connection;
 
 end:
-    g_clear_object(&connection);
     coroutine_yieldto(open_host->from, NULL);
 }
 
@@ -1709,7 +1715,7 @@ static gboolean open_host_idle_cb(gpointer data)
     SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(open_host->session);
 
     g_return_val_if_fail(open_host != NULL, FALSE);
-    g_return_val_if_fail(open_host->socket == NULL, FALSE);
+    g_return_val_if_fail(open_host->connection == NULL, FALSE);
 
 #if GLIB_CHECK_VERSION(2,26,0)
     open_host->proxy = s->proxy;
@@ -1740,20 +1746,28 @@ static gboolean open_host_idle_cb(gpointer data)
 
 /* coroutine context */
 G_GNUC_INTERNAL
-GSocket* spice_session_channel_open_host(SpiceSession *session, SpiceChannel *channel,
-                                         gboolean use_tls)
+GSocketConnection* spice_session_channel_open_host(SpiceSession *session, SpiceChannel *channel,
+                                                   gboolean use_tls)
 {
     SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     spice_open_host open_host = { 0, };
-
-    if ((use_tls && !s->tls_port) || (!use_tls && !s->port))
-        return NULL;
+    gchar *port, *endptr;
 
     // FIXME: make open_host() cancellable
     open_host.from = coroutine_self();
     open_host.session = session;
     open_host.channel = channel;
-    open_host.port = atoi(use_tls ? s->tls_port : s->port);
+    port = use_tls ? s->tls_port : s->port;
+    if (port == NULL)
+        return NULL;
+
+    open_host.port = strtol(port, &endptr, 10);
+    if (*port == '\0' || *endptr != '\0' ||
+        open_host.port <= 0 || open_host.port > G_MAXUINT16) {
+        g_warning("Invalid port value %s", port);
+        return NULL;
+    }
+
     open_host.client = g_socket_client_new();
 
     guint id = g_idle_add(open_host_idle_cb, &open_host);
@@ -1764,13 +1778,15 @@ GSocket* spice_session_channel_open_host(SpiceSession *session, SpiceChannel *ch
     if (open_host.error != NULL) {
         g_warning("%s", open_host.error->message);
         g_clear_error(&open_host.error);
-    } else if (open_host.socket != NULL) {
-        g_socket_set_blocking(open_host.socket, FALSE);
-        g_socket_set_keepalive(open_host.socket, TRUE);
+    } else if (open_host.connection != NULL) {
+        GSocket *socket;
+        socket = g_socket_connection_get_socket(open_host.connection);
+        g_socket_set_blocking(socket, FALSE);
+        g_socket_set_keepalive(socket, TRUE);
     }
 
     g_clear_object(&open_host.client);
-    return open_host.socket;
+    return open_host.connection;
 }
 
 
