@@ -314,8 +314,8 @@ static void update_monitor_area(SpiceDisplay *display)
     }
 
     if (!d->resize_guest_enable)
-        spice_main_set_display(d->main, get_display_id(display),
-                               c->x, c->y, c->width, c->height);
+        spice_main_update_display(d->main, get_display_id(display),
+                                  c->x, c->y, c->width, c->height, FALSE);
 
     update_area(display, c->x, c->y, c->width, c->height);
     g_clear_pointer(&monitors, g_array_unref);
@@ -477,17 +477,18 @@ static GdkCursor* get_blank_cursor(void)
 static gboolean grab_broken(SpiceDisplay *self, GdkEventGrabBroken *event,
                             gpointer user_data G_GNUC_UNUSED)
 {
-    SpiceDisplayPrivate *d = self->priv;
+    SPICE_DEBUG("%s (implicit: %d, keyboard: %d)", __FUNCTION__,
+                event->implicit, event->keyboard);
 
-    SPICE_DEBUG("%s (%d)", __FUNCTION__, event->implicit);
     if (event->keyboard) {
-        d->keyboard_grab_active = false;
-        g_signal_emit(self, signals[SPICE_DISPLAY_KEYBOARD_GRAB], 0, false);
-    } else {
-        d->mouse_grab_active = false;
-        g_signal_emit(self, signals[SPICE_DISPLAY_MOUSE_GRAB], 0, false);
+        try_keyboard_ungrab(self);
     }
 
+    /* always release mouse when grab broken, this could be more
+       generally placed in keyboard_ungrab(), but one might worry of
+       breaking someone else code. */
+    try_mouse_ungrab(self);
+
     return false;
 }
 
@@ -536,7 +537,7 @@ static void spice_display_init(SpiceDisplay *display)
 {
     GtkWidget *widget = GTK_WIDGET(display);
     SpiceDisplayPrivate *d;
-    GtkTargetEntry targets = {"text/plain", 0, 0};
+    GtkTargetEntry targets = { "text/uri-list", 0, 0 };
 
     d = display->priv = SPICE_DISPLAY_GET_PRIVATE(display);
 
@@ -1277,6 +1278,14 @@ static gboolean key_event(GtkWidget *widget, GdkEventKey *key)
     /* on windows, we ought to ignore the reserved key event? */
     if (key->hardware_keycode == 0xff)
         return false;
+
+    if (!d->keyboard_grab_active) {
+        if (key->hardware_keycode == VK_LWIN ||
+            key->hardware_keycode == VK_RWIN ||
+            key->hardware_keycode == VK_APPS)
+            return false;
+    }
+
 #endif
     SPICE_DEBUG("%s %s: keycode: %d  state: %d  group %d modifier %d",
             __FUNCTION__, key->type == GDK_KEY_PRESS ? "press" : "release",
@@ -1654,14 +1663,26 @@ static gboolean configure_event(GtkWidget *widget, GdkEventConfigure *conf)
     SpiceDisplay *display = SPICE_DISPLAY(widget);
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
 
+    if (conf->width == d->ww && conf->height == d->wh &&
+            conf->x == d->mx && conf->y == d->my) {
+        return true;
+    }
+
     if (conf->width != d->ww  || conf->height != d->wh) {
         d->ww = conf->width;
         d->wh = conf->height;
         recalc_geometry(widget);
     }
 
-    try_mouse_ungrab(display);
-    try_mouse_grab(display);
+    d->mx = conf->x;
+    d->my = conf->y;
+
+#ifdef WIN32
+    if (d->mouse_grab_active) {
+        try_mouse_ungrab(display);
+        try_mouse_grab(display);
+    }
+#endif
 
     return true;
 }
