@@ -50,6 +50,12 @@
 
 #include <glib/gi18n.h>
 
+#ifndef G_OS_WIN32 /* Linux -- device id is bus.addr */
+#define DEV_ID_FMT "at %d.%d"
+#else /* Windows -- device id is vid:pid */
+#define DEV_ID_FMT "0x%04x:0x%04x"
+#endif
+
 /**
  * SECTION:usb-device-manager
  * @short_description: USB device management
@@ -597,8 +603,13 @@ static gboolean spice_usb_device_manager_get_udev_bus_n_address(
 
     *bus = *address = 0;
 
+#ifndef G_OS_WIN32
     bus_str = g_udev_device_get_property(udev, "BUSNUM");
     address_str = g_udev_device_get_property(udev, "DEVNUM");
+#else /* Windows -- request vid:pid instead */
+    bus_str = g_udev_device_get_property(udev, "VID");
+    address_str = g_udev_device_get_property(udev, "PID");
+#endif
     if (bus_str)
         *bus = atoi(bus_str);
     if (address_str)
@@ -607,6 +618,50 @@ static gboolean spice_usb_device_manager_get_udev_bus_n_address(
     return *bus && *address;
 }
 
+static gboolean spice_usb_device_manager_get_device_descriptor(
+    libusb_device *libdev,
+    struct libusb_device_descriptor *desc)
+{
+    int errcode;
+    const gchar *errstr;
+
+    g_return_val_if_fail(libdev != NULL, FALSE);
+    g_return_val_if_fail(desc   != NULL, FALSE);
+
+    errcode = libusb_get_device_descriptor(libdev, desc);
+    if (errcode < 0) {
+        int bus, addr;
+
+        bus = libusb_get_bus_number(libdev);
+        addr = libusb_get_device_address(libdev);
+        errstr = spice_usbutil_libusb_strerror(errcode);
+        g_warning("cannot get device descriptor for (%p) %d.%d -- %s(%d)",
+                  libdev, bus, addr, errstr, errcode);
+        return FALSE;
+    }
+    return TRUE;
+}
+
+static gboolean spice_usb_device_manager_get_libdev_vid_pid(
+    libusb_device *libdev, int *vid, int *pid)
+{
+    struct libusb_device_descriptor desc;
+
+    g_return_val_if_fail(libdev != NULL, FALSE);
+    g_return_val_if_fail(vid != NULL, FALSE);
+    g_return_val_if_fail(pid != NULL, FALSE);
+
+    *vid = *pid = 0;
+
+    if (!spice_usb_device_manager_get_device_descriptor(libdev, &desc)) {
+        return FALSE;
+    }
+    *vid = desc.idVendor;
+    *pid = desc.idProduct;
+
+    return TRUE;
+}
+
 /* ------------------------------------------------------------------ */
 /* callbacks                                                          */
 
@@ -655,9 +710,48 @@ static void spice_usb_device_manager_auto_connect_cb(GObject      *gobject,
     spice_usb_device_unref(device);
 }
 
+#ifndef G_OS_WIN32 /* match functions for Linux -- match by bus.addr */
+static gboolean
+spice_usb_device_manager_device_match(SpiceUsbDevice *device,
+                                      const int bus, const int address)
+{
+    return (spice_usb_device_get_busnum(device) == bus &&
+            spice_usb_device_get_devaddr(device) == address);
+}
+
+static gboolean
+spice_usb_device_manager_libdev_match(libusb_device *libdev,
+                                      const int bus, const int address)
+{
+    return (libusb_get_bus_number(libdev) == bus &&
+            libusb_get_device_address(libdev) == address);
+}
+
+#else /* Win32 -- match functions for Windows -- match by vid:pid */
+static gboolean
+spice_usb_device_manager_device_match(SpiceUsbDevice *device,
+                                      const int vid, const int pid)
+{
+    return (spice_usb_device_get_vid(device) == vid &&
+            spice_usb_device_get_pid(device) == pid);
+}
+
+static gboolean
+spice_usb_device_manager_libdev_match(libusb_device *libdev,
+                                      const int vid, const int pid)
+{
+    int vid2, pid2;
+
+    if (!spice_usb_device_manager_get_libdev_vid_pid(libdev, &vid2, &pid2)) {
+        return FALSE;
+    }
+    return (vid == vid2 && pid == pid2);
+}
+#endif /* of Win32 -- match functions */
+
 static SpiceUsbDevice*
 spice_usb_device_manager_find_device(SpiceUsbDeviceManager *self,
-                                     guint8 bus, guint8 address)
+                                     const int bus, const int address)
 {
     SpiceUsbDeviceManagerPrivate *priv = self->priv;
     SpiceUsbDevice *curr, *device = NULL;
@@ -665,8 +759,7 @@ spice_usb_device_manager_find_device(SpiceUsbDeviceManager *self,
 
     for (i = 0; i < priv->devices->len; i++) {
         curr = g_ptr_array_index(priv->devices, i);
-        if (spice_usb_device_get_busnum(curr) == bus &&
-               spice_usb_device_get_devaddr(curr) == address) {
+        if (spice_usb_device_manager_device_match(curr, bus, address)) {
             device = curr;
             break;
         }
@@ -701,7 +794,11 @@ static void spice_usb_device_manager_add_dev(SpiceUsbDeviceManager  *self,
 
     device = spice_usb_device_manager_find_device(self, bus, address);
     if (device) {
-        SPICE_DEBUG("USB device at %d.%d already exists, ignored", bus, address);
+        SPICE_DEBUG("USB device 0x%04x:0x%04x at %d.%d already exists, ignored",
+                    spice_usb_device_get_vid(device),
+                    spice_usb_device_get_pid(device),
+                    spice_usb_device_get_busnum(device),
+                    spice_usb_device_get_devaddr(device));
         return;
     }
 
@@ -711,8 +808,7 @@ static void spice_usb_device_manager_add_dev(SpiceUsbDeviceManager  *self,
         libusb_get_device_list(priv->context, &dev_list);
 
     for (i = 0; dev_list && dev_list[i]; i++) {
-        if (libusb_get_bus_number(dev_list[i]) == bus &&
-            libusb_get_device_address(dev_list[i]) == address) {
+        if (spice_usb_device_manager_libdev_match(dev_list[i], bus, address)) {
             libdev = dev_list[i];
             break;
         }
@@ -732,7 +828,7 @@ static void spice_usb_device_manager_add_dev(SpiceUsbDeviceManager  *self,
         libusb_free_device_list(dev_list, 1);
 
     if (!device) {
-        g_warning("Could not find USB device to add at busnum %d devaddr %d",
+        g_warning("Could not find USB device to add " DEV_ID_FMT,
                   bus, address);
         return;
     }
@@ -768,7 +864,7 @@ static void spice_usb_device_manager_remove_dev(SpiceUsbDeviceManager  *self,
 
     device = spice_usb_device_manager_find_device(self, bus, address);
     if (!device) {
-        g_warning("Could not find USB device to remove at busnum %d devaddr %d",
+        g_warning("Could not find USB device to remove " DEV_ID_FMT,
                   bus, address);
         return;
     }
@@ -777,7 +873,7 @@ static void spice_usb_device_manager_remove_dev(SpiceUsbDeviceManager  *self,
     const guint8 state = spice_usb_device_get_state(device);
     if ((state == SPICE_USB_DEVICE_STATE_INSTALLING) ||
         (state == SPICE_USB_DEVICE_STATE_UNINSTALLING)) {
-        SPICE_DEBUG("skipping device at %d.%d. It is un/installing it's driver",
+        SPICE_DEBUG("skipping " DEV_ID_FMT ". It is un/installing its driver",
                     bus, address);
         return;
     }
@@ -786,8 +882,10 @@ static void spice_usb_device_manager_remove_dev(SpiceUsbDeviceManager  *self,
     spice_usb_device_manager_disconnect_device(self, device);
 
     SPICE_DEBUG("device removed %p", device);
-    g_signal_emit(self, signals[DEVICE_REMOVED], 0, device);
+    spice_usb_device_ref(device);
     g_ptr_array_remove(priv->devices, device);
+    g_signal_emit(self, signals[DEVICE_REMOVED], 0, device);
+    spice_usb_device_unref(device);
 }
 
 static void spice_usb_device_manager_uevent_cb(GUdevClient     *client,
@@ -1170,8 +1268,10 @@ _spice_usb_device_manager_connect_device_async(SpiceUsbDeviceManager *self,
              * So remove the device now
              */
             SPICE_DEBUG("libdev does not exist for %p -- removing", device);
-            g_signal_emit(self, signals[DEVICE_REMOVED], 0, device);
+            spice_usb_device_ref(device);
             g_ptr_array_remove(priv->devices, device);
+            g_signal_emit(self, signals[DEVICE_REMOVED], 0, device);
+            spice_usb_device_unref(device);
 #endif
             g_simple_async_result_set_error(result,
                                             SPICE_CLIENT_ERROR,
@@ -1444,9 +1544,7 @@ gchar *spice_usb_device_get_description(SpiceUsbDevice *device, const gchar *for
 static SpiceUsbDeviceInfo *spice_usb_device_new(libusb_device *libdev)
 {
     SpiceUsbDeviceInfo *info;
-    struct libusb_device_descriptor desc;
-    int errcode;
-    const gchar *errstr;
+    int vid, pid;
     guint8 bus, addr;
 
     g_return_val_if_fail(libdev != NULL, NULL);
@@ -1454,11 +1552,7 @@ static SpiceUsbDeviceInfo *spice_usb_device_new(libusb_device *libdev)
     bus = libusb_get_bus_number(libdev);
     addr = libusb_get_device_address(libdev);
 
-    errcode = libusb_get_device_descriptor(libdev, &desc);
-    if (errcode < 0) {
-        errstr = spice_usbutil_libusb_strerror(errcode);
-        g_warning("cannot get device descriptor for (%p) %d.%d -- %s(%d)",
-                  libdev, bus, addr, errstr, errcode);
+    if (!spice_usb_device_manager_get_libdev_vid_pid(libdev, &vid, &pid)) {
         return NULL;
     }
 
@@ -1466,43 +1560,43 @@ static SpiceUsbDeviceInfo *spice_usb_device_new(libusb_device *libdev)
 
     info->busnum  = bus;
     info->devaddr = addr;
-    info->vid = desc.idVendor;
-    info->pid = desc.idProduct;
+    info->vid = vid;
+    info->pid = pid;
     info->ref = 1;
 
     return info;
 }
 
-guint8 spice_usb_device_get_busnum(SpiceUsbDevice *device)
+guint8 spice_usb_device_get_busnum(const SpiceUsbDevice *device)
 {
-    SpiceUsbDeviceInfo *info = (SpiceUsbDeviceInfo *)device;
+    const SpiceUsbDeviceInfo *info = (const SpiceUsbDeviceInfo *)device;
 
     g_return_val_if_fail(info != NULL, 0);
 
     return info->busnum;
 }
 
-guint8 spice_usb_device_get_devaddr(SpiceUsbDevice *device)
+guint8 spice_usb_device_get_devaddr(const SpiceUsbDevice *device)
 {
-    SpiceUsbDeviceInfo *info = (SpiceUsbDeviceInfo *)device;
+    const SpiceUsbDeviceInfo *info = (const SpiceUsbDeviceInfo *)device;
 
     g_return_val_if_fail(info != NULL, 0);
 
     return info->devaddr;
 }
 
-guint16 spice_usb_device_get_vid(SpiceUsbDevice *device)
+guint16 spice_usb_device_get_vid(const SpiceUsbDevice *device)
 {
-    SpiceUsbDeviceInfo *info = (SpiceUsbDeviceInfo *)device;
+    const SpiceUsbDeviceInfo *info = (const SpiceUsbDeviceInfo *)device;
 
     g_return_val_if_fail(info != NULL, 0);
 
     return info->vid;
 }
 
-guint16 spice_usb_device_get_pid(SpiceUsbDevice *device)
+guint16 spice_usb_device_get_pid(const SpiceUsbDevice *device)
 {
-    SpiceUsbDeviceInfo *info = (SpiceUsbDeviceInfo *)device;
+    const SpiceUsbDeviceInfo *info = (const SpiceUsbDeviceInfo *)device;
 
     g_return_val_if_fail(info != NULL, 0);
 
@@ -1552,6 +1646,7 @@ static void spice_usb_device_unref(SpiceUsbDevice *device)
     }
 }
 
+#ifndef G_OS_WIN32 /* Linux -- compare bus.addr */
 static gboolean
 spice_usb_device_equal_libdev(SpiceUsbDevice *device,
                               libusb_device  *libdev)
@@ -1568,6 +1663,26 @@ spice_usb_device_equal_libdev(SpiceUsbDevice *device,
 
     return ((bus1 == bus2) && (addr1 == addr2));
 }
+#else /* Windows -- compare vid:pid of device and libdev */
+static gboolean
+spice_usb_device_equal_libdev(SpiceUsbDevice *device,
+                              libusb_device  *libdev)
+{
+    int vid1, vid2, pid1, pid2;
+
+    if ((device == NULL) || (libdev == NULL))
+        return FALSE;
+
+    vid1 = spice_usb_device_get_vid(device);
+    pid1 = spice_usb_device_get_pid(device);
+
+    if (!spice_usb_device_manager_get_libdev_vid_pid(libdev, &vid2, &pid2)) {
+        return FALSE;
+    }
+
+    return ((vid1 == vid2) && (pid1 == pid2));
+}
+#endif
 
 /*
  * Caller must libusb_unref_device the libusb_device returned by this function.
@@ -1578,7 +1693,7 @@ spice_usb_device_manager_device_to_libdev(SpiceUsbDeviceManager *self,
                                           SpiceUsbDevice *device)
 {
     libusb_device *d, **devlist;
-    guint8 bus, addr;
+    int bus, addr;
     int i;
 
     g_return_val_if_fail(SPICE_IS_USB_DEVICE_MANAGER(self), NULL);
@@ -1586,16 +1701,20 @@ spice_usb_device_manager_device_to_libdev(SpiceUsbDeviceManager *self,
     g_return_val_if_fail(self->priv != NULL, NULL);
     g_return_val_if_fail(self->priv->context != NULL, NULL);
 
+#ifndef G_OS_WIN32
     bus  = spice_usb_device_get_busnum(device);
     addr = spice_usb_device_get_devaddr(device);
+#else
+    bus  = spice_usb_device_get_vid(device);
+    addr = spice_usb_device_get_pid(device);
+#endif
 
     libusb_get_device_list(self->priv->context, &devlist);
     if (!devlist)
         return NULL;
 
     for (i = 0; (d = devlist[i]) != NULL; i++) {
-        if ((libusb_get_bus_number(d) == bus) &&
-            (libusb_get_device_address(d) == addr)) {
+        if (spice_usb_device_manager_libdev_match(d, bus, addr)) {
             libusb_ref_device(d);
             break;
         }
